/*
 * Project: MoleCuilder
 * Description: creates and alters molecular systems
 * Copyright (C)  2010-2012 University of Bonn. All rights reserved.
 * 
 *
 *   This file is part of MoleCuilder.
 *
 *    MoleCuilder is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    MoleCuilder is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with MoleCuilder.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * AddAction.cpp
 *
 *  Created on: May 9, 2010
 *      Author: heber
 */

// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "CodePatterns/MemDebug.hpp"

#include "Descriptors/AtomIdDescriptor.hpp"
#include "Atom/atom.hpp"
#include "Element/element.hpp"
#include "CodePatterns/Log.hpp"
#include "molecule.hpp"
#include "LinearAlgebra/Vector.hpp"
#include "CodePatterns/Verbose.hpp"
#include "World.hpp"

#include <iostream>
#include <string>

#include "Actions/AtomAction/AddAction.hpp"

using namespace MoleCuilder;

// and construct the stuff
#include "AddAction.def"
#include "Action_impl_pre.hpp"
/** =========== define the function ====================== */

atom * getNewAtom(const AtomAddAction::AtomAddParameters &_params)
{
  atom * first = World::getInstance().createAtom();
  first->setType(_params.elemental.get());
  first->setPosition(_params.position.get());

  return first;
}

std::vector<atomId_t> createAtoms(
    const AtomAddAction::AtomAddParameters &_params,
    std::vector<molecule *> &_molecules)
{
  std::vector<atomId_t> ids;
  if (!_molecules.empty()) {
    if (_molecules.size() == 1) {
        atom *first = getNewAtom(_params);
        molecule *mol = *_molecules.begin();
        LOG(1, "Adding new atom with element " << first->getType()->getName()
            << " at " << (first->getPosition()) << " to selected molecule "
            << mol->getName()+".");
        mol->AddAtom(first);
      ids.push_back(first->getId());
    }
  } else {
    atom *first = getNewAtom(_params);
    molecule *mol = World::getInstance().createMolecule();
    mol->setName("none");
    mol->AddAtom(first);
    LOG(1, "Adding new atom with element " << first->getType()->getName()
        << " at " << (first->getPosition()) << " to new molecule.");
    ids.push_back(first->getId());
  }

  return ids;
}

ActionState::ptr AtomAddAction::performCall() {
  // execute action
  std::vector<molecule *> molecules = World::getInstance().getSelectedMolecules();
  std::vector<atomId_t> ids = createAtoms(params, molecules);

  if (molecules.size() > 1)
     return Action::failure;
  else
    return ActionState::ptr(new AtomAddState(ids, params));
}

ActionState::ptr AtomAddAction::performUndo(ActionState::ptr _state) {
  AtomAddState *state = assert_cast<AtomAddState*>(_state.get());

  for (std::vector<atomId_t>::const_iterator iter = state->ids.begin();
      iter != state->ids.end(); ++iter) {
    LOG(1, "Removing atom with id " << *iter << ".");
    World::getInstance().destroyAtom(*iter);
  }

  return ActionState::ptr(_state);
}

ActionState::ptr AtomAddAction::performRedo(ActionState::ptr _state){
  AtomAddState *state = assert_cast<AtomAddState*>(_state.get());

  std::vector<molecule *> molecules = World::getInstance().getSelectedMolecules();
  std::vector<atomId_t> newids = createAtoms(params, molecules);

  if (newids.size() != state->ids.size()) {
    STATUS("Could not recreate all atoms after undo.");
    for (std::vector<atomId_t>::const_iterator iter = newids.begin(); iter != newids.end();++iter)
      World::getInstance().destroyAtom(*iter);
    return Action::failure;
  }

  std::vector<atomId_t>::const_iterator newiter = newids.begin();
  std::vector<atomId_t>::const_iterator olditer = state->ids.begin();
  bool status=true;
  for (; newiter != newids.end(); ++newiter, ++olditer) {
    atom * first = World::getInstance().getAtom(AtomById(*newiter));
    ASSERT( first != NULL,
        "AtomAddAction::performRedo() - re-created atom not present?");
    if (first->getId() != *olditer)
      if (!first->changeId(*olditer)) {
        STATUS("Could not change atom id "+toString(first->getId())+"->"+toString(*olditer)+".");
        // remove all created atoms
        for (std::vector<atomId_t>::const_iterator iter = state->ids.begin(); iter != olditer;++iter)
          World::getInstance().destroyAtom(*iter);
        olditer = state->ids.end();
        for (std::vector<atomId_t>::const_iterator iter = newiter; iter != newids.end();++iter)
          World::getInstance().destroyAtom(*iter);
        status = false;
        break;
      }
  }
  ASSERT( olditer == state->ids.end(),
      "AtomAddAction::performRedo() - after all unequal amount of ids in new and old?");

  if (!status) {
    return Action::failure;
  } else
    return ActionState::ptr(_state);
}

bool AtomAddAction::canUndo() {
  return true;
}

bool AtomAddAction::shouldUndo() {
  return true;
}
/** =========== end of function ====================== */
